import mss
from PIL import Image
from quack2tex.pyqt import QLayout, QWidget, QApplication, QRect, QSize, QObject, QMessageBox


class GuiUtils:
    """
    Utility functions for the GUI.
    """

    @classmethod
    def clear_layout(cls, layout: QLayout):
        """
        clear a pyqt layout
        """
        while layout.count():
            child = layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()


    @staticmethod
    def get_current_monitor_index(widget: QWidget) -> int:
        """
        Get the index of the monitor where the widget is located.
        :param widget: The widget whose monitor index is needed.
        :return: Index of the monitor containing the widget.
        """
        window_geometry = widget.frameGeometry()
        screens = QApplication.screens()
        for i, screen in enumerate(screens):
            if screen.geometry().contains(window_geometry.topLeft()):
                return i
        return -1  # Return -1 if no monitor is found (edge case).

    @staticmethod
    def get_current_monitor_geometry(widget: QWidget) -> QRect:
        """
        Get the geometry of the monitor where the widget is located.
        :param widget: The widget whose monitor geometry is needed.
        :return: Geometry of the monitor containing the widget.
        """
        current_monitor_index = GuiUtils.get_current_monitor_index(widget)
        return QApplication.screens()[current_monitor_index].availableGeometry()


    @staticmethod
    def get_screen_capture_image(screen_region, monitor_index):
        """
        Capture the screen region
        :param screen_region:
        :param monitor_index:
        :return:
        """
        # TODO: Add support for multiple monitors and DPI scaling factors

        with mss.mss() as sct:
            monitor = {
                "top": screen_region[1],
                "left": screen_region[0],
                "width": screen_region[2],
                "height": screen_region[3],
            }
            sct_img = sct.grab(monitor)
            img = Image.frombytes("RGB", sct_img.size, sct_img.bgra, "raw", "BGRX")
            img.format = "PNG"
            return img

    @staticmethod
    def move_window_to_center(window: QWidget):
        """
        Move the window to the center of the screen.
        :param window: The window to move.
        """
        screen = QApplication.primaryScreen()
        screen_geometry = screen.availableGeometry()
        x = screen_geometry.width() // 2 - window.width() // 2
        y = screen_geometry.height() // 2 - window.height() // 2
        window.move(x, y)

    @staticmethod
    def move_window_to_top_center(widget: QWidget):
        """
        Move the window to the top center of the screen.
        :param widget: The widget to move.
        """
        screen = QApplication.primaryScreen()
        screen_geometry = screen.availableGeometry()
        x = screen_geometry.width() // 2 - widget.width() // 2
        y = screen_geometry.top()
        widget.move(x, y)

    @classmethod
    def move_widget_to_center(cls, widget: QWidget):
        """
        Center the widget on its current screen.
        :param widget: The widget to move.
        """
        screen_geometry = cls.get_current_monitor_geometry(widget)
        x = screen_geometry.center().x() - widget.width() // 2
        y = screen_geometry.center().y() - widget.height() // 2
        widget.move(x, y)

    @classmethod
    def move_widget_to_widget_bottom(cls, widget: QWidget, parent_widget: QWidget):
        """
        Move the widget to the bottom center of the parent widget.
        :param widget: The widget to move.
        :param parent_widget: The parent widget used as reference.
        """
        x = parent_widget.x() + (parent_widget.width() - widget.width()) // 2
        y = parent_widget.y() + parent_widget.height() - widget.height()
        widget.move(x, y)

    @staticmethod
    def calculate_new_size(image_sz: QSize, target_sz: QSize) -> QSize:
        """
        Calculate the new size of an image to fit into a target rectangle while preserving the aspect ratio.
        :param image_sz: The size of the image.
        :param target_sz: The size of the target rectangle.
        :return: The new QSize with the aspect ratio preserved.
        """
        image_aspect_ratio = image_sz.width() / image_sz.height()
        target_aspect_ratio = target_sz.width() / target_sz.height()

        if image_aspect_ratio > target_aspect_ratio:
            new_width = target_sz.width()
            new_height = int(new_width / image_aspect_ratio)
        else:
            new_height = target_sz.height()
            new_width = int(new_height * image_aspect_ratio)

        return QSize(new_width, new_height)

    @classmethod
    def move_widget_to_center_top(cls, widget: QWidget):
        """
        Center the widget at the top of its current screen.
        :param widget: The widget to move.
        """
        screen_geometry = cls.get_current_monitor_geometry(widget)
        x = screen_geometry.center().x() - widget.width() // 2
        y = screen_geometry.top()
        widget.move(x, y)

    @classmethod
    def move_widget_to_center_bottom(cls, widget: QWidget):
        """
        Center the widget at the bottom of its current screen.
        :param widget: The widget to move.
        """
        screen_geometry = cls.get_current_monitor_geometry(widget)
        x = screen_geometry.center().x() - widget.width() // 2
        y = screen_geometry.bottom() - widget.height()
        widget.move(x, y)



    @staticmethod
    def get_clipboard_text():
        """
        Get the text from the clipboard.
        :return: The text from the clipboard.
        """
        clipboard = QApplication.clipboard()
        return clipboard.text()

    @staticmethod
    def bind(objectName, propertyName, propertyType):
        """
        Bind a property of a widget to a property of the parent widget.
        :param objectName:
        :param propertyName:
        :param propertyType:
        :return:
        """
        def getter(self):
            widget = self.findChild(QObject, objectName)
            prop_value = widget.property(propertyName)
            return propertyType(prop_value)

        def setter(self, value):
            self.findChild(QObject, objectName).setProperty(propertyName, value)

        return property(getter, setter)


    @staticmethod
    def show_error(message: str):
        """
        Show an error message dialog.
        :param message: The error message to display.
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Icon.Critical)
        msg.setText(message)
        msg.setWindowTitle("Error")
        msg.exec()

    @staticmethod
    def show_info(message: str):
        """
        Show an information message dialog.
        :param message: The information message to display.
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Icon.Information)
        msg.setText(message)
        msg.setWindowTitle("Information")
        msg.exec()






import numpy as np
from dotenv import load_dotenv, find_dotenv

import quack2tex

load_dotenv(find_dotenv())



@quack2tex.latify(model="models/gemini-1.5-flash-latest")
def sqrt(x: float):
    """
    Compute the square root of a number.
    :param x: The number.
    :return: The square root of the number.
    """
    x = x ** 0.5
    return x

@quack2tex.latify(model="models/gemini-1.5-flash-latest")
def riemman_sum(f, a, b, n):
    """
    Compute the Riemman sum of a function.
    :param f: The function.
    :param a: The lower bound.
    :param b: The upper bound.
    :param n: The number of subintervals.
    :return: The Riemman sum.
    """
    dx = (b - a) / n
    return sum(f(a + i * dx) * dx for i in range(n))

@quack2tex.latify(model="models/gemini-1.5-flash-latest")
def gaussian_function(x):
   """
   Compute the integral of a function.
   :param x: The number.
   :return: The integral of the number.
   """
   return np.exp(-x**2)

@quack2tex.latify(model="models/gemini-1.5-flash-latest")
def standard_deviation(data):
    """
    Compute the integral of a function.
    :param f: The function.
    :param a: The lower bound.
    :param b: The upper bound.
    :param n: The number of subintervals.
    :return: The integral.
    """
    return np.std(data)


if __name__ == "__main__":
    quack2tex.run_app()



from typing import Any

from quack2tex import LLM
from quack2tex.pyqt import (
    QDialog,
    QVBoxLayout,
    QFormLayout,
    QTextEdit,
    QSizePolicy,
    QDialogButtonBox,
    QComboBox,
    Qt,
    QObject,
    Signal,
    QLabel,
    QThreadPool,
    Property
)
from quack2tex.utils import GuiUtils, Worker, work_exception
from quack2tex.widgets import FileUploader, ModelPicker


class CaptureModeComboBox(QComboBox):
    """
    A combo box for selecting the capture mode
    """

    def __init__(self, parent=None):
        super(CaptureModeComboBox, self).__init__(parent)
        self.addItem("Select Capture Mode", userData=None)
        self.addItem("Screen", userData="screen")
        self.addItem("Clipboard", userData="clipboard")

    def get_capture_mode(self):
        """
        Get the capture mode
        :return:
        """
        return self.currentData(Qt.ItemDataRole.UserRole)  # Corrected role usage

    def set_capture_mode(self, mode):
        """
        Set the capture mode
        :param mode:
        :return:
        """
        for i in range(self.count()):
            if self.itemData(i, Qt.ItemDataRole.UserRole) == mode:  # Corrected role usage
                self.setCurrentIndex(i)
                break

    capture_mode = Property(str, get_capture_mode, set_capture_mode)


class MenuItemForm(QDialog, QObject):
    """
    A custom form for editing menu items
    """
    name = GuiUtils.bind("txt_name", "plainText", str)
    icon = GuiUtils.bind("file_icon_uploader", "file_path", str)
    system_instruction = GuiUtils.bind("txt_system_instructions", "plainText", str)
    guidance_prompt = GuiUtils.bind("txt_guidance_prompt", "plainText", str)
    models = GuiUtils.bind("list_model_picker", "models", str)
    capture_mode = GuiUtils.bind("cbx_capture_mode", "capture_mode", str)

    on_widget_loaded = Signal(dict)

    def __init__(self, parent=None):
        super(MenuItemForm, self).__init__(parent)
        self.setWindowTitle("Edit menu item")
        self.setModal(True)

        # Main Layout
        self.layout = QVBoxLayout(self)
        self.setFixedSize(400, 400)

        self.loading_label = QLabel("Loading data...")
        self.loading_label.setAlignment(Qt.AlignmentFlag.AlignCenter)  # Corrected alignment flag usage
        self.on_widget_loaded.connect(self.on_widget_loaded_handler)
        self.layout.addWidget(self.loading_label)
        self.thread_pool = QThreadPool()
        self.form_data = None
        self.load_data()


    def build_form(self, data: dict):
        """
        Create the form
        :return:
        """
        GuiUtils.clear_layout(self.layout)
        # Form Layout
        form_layout = QFormLayout()
        form_layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        form_layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight)  # Corrected alignment flag usage

        self.txt_name = QTextEdit(self)
        self.txt_name.setFixedHeight(30)
        self.txt_name.setObjectName("txt_name")
        # dont allow rich text
        self.txt_name.setAcceptRichText(False)
        self.txt_name.setPlaceholderText("Enter menu item name...")
        self.txt_name.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)  # Corrected QSizePolicy usage
        form_layout.addRow("Name:", self.txt_name)

        self.file_icon_uploader = FileUploader(self, file_filter="Images (*.png *.jpg *.jpeg)")
        self.file_icon_uploader.setObjectName("file_icon_uploader")
        self.file_icon_uploader.file_path_edit.setPlaceholderText("Select icon file...")
        self.file_icon_uploader.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)  # Corrected QSizePolicy usage
        form_layout.addRow("Icon:", self.file_icon_uploader)

        self.txt_system_instructions = QTextEdit(self)
        self.txt_system_instructions.setObjectName("txt_system_instructions")
        self.txt_system_instructions.setAcceptRichText(False)
        self.txt_system_instructions.setPlaceholderText("Enter system instructions...")
        self.txt_system_instructions.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)  # Corrected QSizePolicy usage
        form_layout.addRow("System Instruction:", self.txt_system_instructions)

        self.txt_guidance_prompt = QTextEdit(self)
        self.txt_guidance_prompt.setObjectName("txt_guidance_prompt")
        self.txt_guidance_prompt.setAcceptRichText(False)
        self.txt_guidance_prompt.setPlaceholderText("Enter guidance prompt...")
        self.txt_guidance_prompt.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)  # Corrected QSizePolicy usage
        form_layout.addRow("Guidance Prompt:", self.txt_guidance_prompt)

        # Model Selection (Multiple Selection)
        self.list_model_picker = ModelPicker(self)
        self.list_model_picker.setObjectName("list_model_picker")
        self.list_model_picker.set_data(data["models"])
        self.list_model_picker.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)  # Corrected QSizePolicy usage
        form_layout.addRow("Select Models:", self.list_model_picker)

        self.cbx_capture_mode = CaptureModeComboBox(self)
        self.cbx_capture_mode.setObjectName("cbx_capture_mode")
        self.cbx_capture_mode.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)  # Corrected QSizePolicy usage
        form_layout.addRow("Capture Mode:", self.cbx_capture_mode)

        # Dialog Buttons (OK and Cancel)
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)  # Corrected enum usage
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        form_layout.addRow(button_box)

        # disable tab in the text fields
        self.txt_name.setTabChangesFocus(True)
        self.txt_system_instructions.setTabChangesFocus(True)
        self.txt_guidance_prompt.setTabChangesFocus(True)

        # set tab order
        self.setTabOrder(self.txt_name, self.file_icon_uploader.file_path_edit)
        self.setTabOrder(self.file_icon_uploader.file_path_edit, self.txt_system_instructions)
        self.setTabOrder(self.txt_system_instructions, self.txt_guidance_prompt)
        self.setTabOrder(self.txt_guidance_prompt, self.list_model_picker)
        self.setTabOrder(self.list_model_picker, self.cbx_capture_mode)
        self.setTabOrder(self.cbx_capture_mode, button_box)
        self.setTabOrder(button_box, self.file_icon_uploader.file_path_edit)

        self.layout.addLayout(form_layout)

    def load_data(self):
        """
        Load the models
        """
        @work_exception
        def do_work()-> dict[str, Any]:
            """
            Load the models
            :return:
            """
            data = {"models": LLM.available_models()}
            return data

        def done(result):
            """
            Handler for when the models are loaded
            """
            data, error = result
            if error:
                GuiUtils.show_error(error)
                return

            self.on_widget_loaded.emit(data)


        worker = Worker(do_work)
        worker.signals.result.connect(done)
        self.thread_pool.start(worker)

    def on_widget_loaded_handler(self, data):
        """
        Handler for when the widget is loaded
        """
        self.build_form(data)
        if self.form_data:
            self.name = self.form_data.get("name")
            self.icon = self.form_data.get("icon")
            self.system_instruction = self.form_data.get("system_instruction")
            self.guidance_prompt = self.form_data.get("guidance_prompt")
            self.models = self.form_data.get("models")
            self.capture_mode = self.form_data.get("capture_mode")

    def accept(self):
        """
        Accept the dialog and close it
        """
        if not self.name:
            GuiUtils.show_error("Name and icon are required.")
            return
        return super(MenuItemForm, self).accept()



from typing import AsyncIterator

from .session_manager import SessionManager
from quack2tex.utils import LibUtils
from sqlalchemy.pool import NullPool
from sqlalchemy.ext.asyncio import AsyncSession

_, db_async_connection_string = LibUtils.get_db_connection_string()
sessionmanager = SessionManager(
    url=db_async_connection_string,
    async_mode=True,
    engine_kwargs={"echo": False, "poolclass": NullPool},
)


async def get_db_session(*args, **kwargs)-> AsyncIterator[AsyncSession]:
    """
    Get database session
    """
    drop_all = kwargs.pop("drop_all", False)
    await sessionmanager.async_init(drop_all=drop_all)
    async with sessionmanager.async_session(*args, **kwargs) as session:
        yield session



import sys

from quack2tex.pyqt import (
    QUrl, QWebChannel, QApplication, QWebEngineView, QObject, Signal, Slot, Property, QWebEnginePage, QWebEngineSettings
)
from quack2tex.resources import *  # noqa: F401


class MarkdownViewerDoc(QObject):
    """
    A simple QObject to expose the markdown viewer to the web page
    """

    contentChanged = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._content = ""

    @Slot(str)
    def log(self, msg):
        """
        Print a message
        """
        #print(msg)
        ...

    @Slot(str)
    def send_to_clipboard(self, text):
        """
        Send text to the clipboard
        """
        clipboard = QApplication.clipboard()
        clipboard.setText(text)

    def get_content(self):
        """
        Get the content of the markdown viewer
        :return:
        """
        return self._content

    def set_content(self, content: str):
        """
        Set the content of the markdown viewer
        :param text:
        :return:
        """
        if self._content == content:
            return
        self._content = content
        self.contentChanged.emit(content)

    content = Property(str, fget=get_content, fset=set_content, notify=contentChanged)


class CustomWebEnginePage(QWebEnginePage):
    """
    A custom web engine page that automatically grants all requested
    """

    def __init__(self, parent=None):
        super().__init__(parent)

    def featurePermissionRequested(self, securityOrigin, feature):
        """
        Automatically grant all requested permissions
        :param securityOrigin:
        :param feature:
        :return:
        """
        # Automatically grant all requested permissions
        self.setFeaturePermission(
            securityOrigin, feature, QWebEnginePage.PermissionPolicy.PermissionGrantedByUser
        )


class MarkdownViewer(QWebEngineView):
    """
    A simple markdown viewer widget
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # Create a web channel
        self.channel = QWebChannel()
        # self.setPage(CustomWebEnginePage(self))
        self.page().setWebChannel(self.channel)
        # Expose the doc object to the web page
        self.doc = MarkdownViewerDoc(self)
        self.channel.registerObject("qtViewerDoc", self.doc)
        # Enable local content access
        self.page().settings().setAttribute(
            QWebEngineSettings.WebAttribute.LocalContentCanAccessRemoteUrls, True
        )
        # Load the index.html file
        #QUrl.fromLocalFile(str(Path(__file__).parent / "files/index.html"))
        local_url = QUrl("qrc:/files/index.html")
        self.load(local_url)

    def set_content(self, content: str):
        """
        Set the content of the markdown viewer
        """
        self.doc.set_content(content)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    viewer = MarkdownViewer()
    viewer.set_content("# Hi\n\nThis is a simple markdown viewer.")
    viewer.show()
    sys.exit(app.exec())



from quack2tex.repository.models import MenuItem
from quack2tex.repository.db.sync_session import get_db_session
from sqlalchemy.orm import Session
from typing import List, Optional


class MenuItemRepository:
    """
    Repository class for handling operations related to MenuItems.
    """

    @staticmethod
    def populate_item_children(session: Session, parent_item: MenuItem) -> None:
        """
        Recursively populates the children of a given menu item.

        Args:
            session (Session): The active database session.
            parent_item (MenuItem): The menu item whose children are to be populated.
        """
        for child in parent_item.children:
            MenuItemRepository.populate_item_children(session, child)

    @classmethod
    def build_tree(cls, session: Session) -> List[MenuItem]:
        """
        Constructs a tree structure of menu items with parent-child relationships.

        This method retrieves all root menu items (items without a parent) and
        recursively appends their children to form a hierarchical structure.

        Args:
            session (Session): The active database session.

        Returns:
            List[MenuItem]: A list of root menu items, each containing its nested children.
        """
        tree: List[MenuItem] = []
        root_items: List[MenuItem] = session.query(MenuItem).filter(MenuItem.parent_id == None).all()

        for root_item in root_items:
            cls.populate_item_children(session, root_item)
            tree.append(root_item)

        return tree

    @classmethod
    def add_item(cls, session: Session, item: MenuItem) -> MenuItem:
        """
        Adds a new menu item to the database.

        Args:
            session (Session): The active database session.
            item (MenuItem): The menu item to add.
            parent_id (Optional[int]): The ID of the parent item, if any.
        """
        if item.parent_id:
            parent_item = session.query(MenuItem).get(item.parent_id)
            parent_item.children.append(item)
        else:
            session.add(item)
        session.commit()
        # Refresh the item to populate its children
        session.refresh(item)
        return item


    @classmethod
    def delete_items(cls, session: Session, item_ids: List[int]) -> None:
        """
        Deletes multiple menu items from the database.

        Args:
            session (Session): The active database session.
            item_ids (List[int]): The IDs of the items to delete.
        """
        items = session.query(MenuItem).filter(MenuItem.id.in_(item_ids)).all()
        for item in items:
            session.delete(item)
        session.commit()

    @classmethod
    def update_item(cls, session: Session, item: MenuItem) -> MenuItem:
        """
        Updates a menu item in the database.

        Args:
            session (Session): The active database session.
            item (MenuItem): The menu item to update.
        """
        item = session.merge(item)
        session.commit()
        return item


if __name__ == '__main__':
    with get_db_session() as session:
        menu_tree = MenuItemRepository.build_tree(session)
        # Optionally, print or use the tree as needed
        print(menu_tree)



from email.policy import default

from numpy.f2py.crackfortran import debug

from quack2tex.pyqt import QSize, QThreadPool
from quack2tex.repository import MenuItemRepository
from quack2tex.repository.db.sync_session import get_db_session
from quack2tex.widgets import FloatingMenu, LoadingIndicator, FloatingMenuItem
from quack2tex.utils.worker import  Worker
from quack2tex.resources import *  # noqa: F401


class DuckMenu(FloatingMenu):
    def __init__(self, parent=None):
        """
        Initialize the DuckMenu with a loading indicator and a thread pool for asynchronous tasks.
        """
        super().__init__(parent=parent)
        self.threadpool = QThreadPool()
        self.loading_indicator = LoadingIndicator(":icons/loading.gif", QSize(200, 100), self)
        self.loading_indicator.hide()
        self.default_icon_path = ":icons/ai.png"

    def clear_menu(self):
        """
        Clear the menu by deleting all FloatingMenuItem children.
        """
        for item in self.findChildren(FloatingMenuItem):
            item.deleteLater()

    def build_menu(self):
        """
        Start building the menu asynchronously by fetching data from the database.
        """
        worker = Worker(self.do_build_menu)
        worker.signals.result.connect(self.done_build_menu)
        self.threadpool.start(worker)

    def do_build_menu(self):
        """
        Synchronously fetch the menu data from the database.
        """
        with get_db_session() as session:
            return MenuItemRepository.build_tree(session)

    def done_build_menu(self, tree):
        """
        Process the fetched menu data and populate the FloatingMenu with items.
        :param tree: The hierarchical menu tree fetched from the database.
        """
        self.clear_menu()
        root_item = self.create_root_item()
        self.add_default_items(root_item)
        self.populate_menu(tree, root_item)
        self.set_root(root_item)
        self.draw_menu()

        self.loading_indicator.move(
            int(self.geometry().center().x() - self.loading_indicator.width() // 2),
            int(self.geometry().center().y())
        )


    def create_root_item(self):
        """
        Create and return the root FloatingMenuItem for the DuckMenu.
        """
        return FloatingMenuItem(
            ":icons/rubber-duck.png",
            QSize(64, 64),
            distance_to_center=100,
            start_angle=0,
            end_angle=360,
            root=None,
            parent=self
        )

    def add_default_items(self, root_item):
        """
        Add default items (e.g., Close and Settings) to the root item.
        :param root_item: The root FloatingMenuItem.
        """
        default_items = [
            {"icon": ":icons/close.png", "tooltip": "Close", "data": {"action": "exit"}},
            {"icon": ":icons/gears.png", "tooltip": "Settings", "data": {"action": "settings"}},
        ]
        for item in default_items:
            self.create_default_item_menu(root_item, item)

    def create_default_item_menu(self, parent_item, action):
        """
        Create a menu item and attach it to a parent item.
        :param parent_item: The parent FloatingMenuItem.
        :param action: Dictionary containing icon, tooltip, and data for the item.
        """
        item = FloatingMenuItem(
            action["icon"] or self.default_icon_path,
            QSize(parent_item.size().width() // 2, parent_item.size().height() // 2),
            distance_to_center=parent_item.distance_to_center // 2,
            start_angle=0,
            end_angle=360,
            parent=self,
            data=action["data"]
        )
        item.setToolTip(action["tooltip"])
        parent_item.add_child(item)

    def populate_menu(self, tree, root_item):
        """
        Recursively populate the menu from the fetched tree.
        :param tree: List of database items representing the menu tree.
        :param root_item: The root FloatingMenuItem.
        """
        for db_item in tree:
            self.add_item_menu_recursive(db_item, root_item)

    def add_item_menu_recursive(self, db_item, parent_item):
        """
        Recursively add items to the menu based on the database item hierarchy.
        :param db_item: Database item containing child items.
        :param parent_item: The parent FloatingMenuItem to which children will be added.
        """
        child_item = FloatingMenuItem(
            db_item.icon or self.default_icon_path,
            QSize(32, 32),
            distance_to_center=parent_item.distance_to_center // 2,
            start_angle=0,
            end_angle=360,
            parent=self
        )
        prompt_node = db_item.guidance_prompt and db_item.system_instruction
        if prompt_node:
            child_item.data = {"action": db_item.name, "tag": db_item}
        child_item.setToolTip(db_item.name)
        parent_item.add_child(child_item)

        for db_child_item in db_item.children:
            self.add_item_menu_recursive(db_child_item, child_item)



from pathlib import Path
from typing import Optional, List

from PIL import Image
from PIL.Image import Image as PILImage

from quack2tex.pyqt import (
    QModelIndex, QThreadPool, QSize, Signal, QIcon, QStandardItem, QDialog, QVBoxLayout, QPushButton, QToolBar,
    QApplication, QTreeView
)
from quack2tex.repository import MenuItemRepository
from quack2tex.repository.db.sync_session import get_db_session
from quack2tex.repository.models import MenuItem
from quack2tex.resources import *  # noqa
from quack2tex.utils import TreeViewStandardItemModel  # Replace CustomModel with this
from quack2tex.utils import Worker, LibUtils, work_exception, GuiUtils
from quack2tex.widgets.forms import MenuItemForm
import sys


class SettingsWindow(QDialog):
    """
    Main application window.
    """
    on_settings_changed = Signal()

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setFixedSize(800, 600)
        self.setWindowTitle("Settings")

        self.layout = QVBoxLayout(self)
        self.setLayout(self.layout)

        self.toolbar = QToolBar()
        self.treeview = QTreeView()

        self.layout.addWidget(self.toolbar)
        self.layout.addWidget(self.treeview)

        self.btn_check_all = QPushButton()
        self.btn_check_all.setIcon(QIcon(":/icons/check-all.png"))
        self.btn_check_all.setIconSize(QSize(16, 16))
        self.btn_check_all.clicked.connect(self.check_all_items)
        self.btn_check_all.setToolTip("Check all items")
        self.toolbar.addWidget(self.btn_check_all)

        self.btn_uncheck_all = QPushButton()
        self.btn_uncheck_all.setIcon(QIcon(":/icons/uncheck-all.png"))
        self.btn_uncheck_all.setIconSize(QSize(16, 16))
        self.btn_uncheck_all.clicked.connect(self.uncheck_all_items)
        self.btn_uncheck_all.setToolTip("Uncheck all items")
        self.toolbar.addWidget(self.btn_uncheck_all)

        self.btn_refresh = QPushButton()
        self.btn_refresh.setIcon(QIcon(":/icons/refresh.png"))
        self.btn_refresh.setIconSize(QSize(16, 16))
        self.btn_refresh.clicked.connect(self.populate_treeview)
        self.btn_refresh.setToolTip("Refresh items")
        self.toolbar.addWidget(self.btn_refresh)

        self.btn_clear_selection = QPushButton()
        self.btn_clear_selection.setIconSize(QSize(16, 16))
        self.btn_clear_selection.setIcon(QIcon(":/icons/broom.png"))
        self.btn_clear_selection.setToolTip("Clear items selection")
        self.btn_clear_selection.clicked.connect(self.treeview.clearSelection)
        self.toolbar.addWidget(self.btn_clear_selection)

        self.toolbar.addSeparator()

        self.btn_add = QPushButton()
        self.btn_add.setIcon(QIcon(":/icons/add.png"))
        self.btn_add.setIconSize(QSize(16, 16))
        self.btn_add.setToolTip("Add new item")
        self.btn_add.clicked.connect(self.add_new_item)
        self.toolbar.addWidget(self.btn_add)

        self.btn_edit = QPushButton()
        self.btn_edit.setIcon(QIcon(":/icons/edit.png"))
        self.btn_edit.setIconSize(QSize(16, 16))
        self.btn_edit.clicked.connect(self.edit_selected_item)
        self.btn_edit.setToolTip("Edit selected item")
        self.toolbar.addWidget(self.btn_edit)

        self.btn_delete = QPushButton()
        self.btn_delete.setIcon(QIcon(":/icons/delete.png"))
        self.btn_delete.setIconSize(QSize(16, 16))
        self.btn_delete.clicked.connect(self.delete_selected_items)
        self.btn_delete.setToolTip("Delete checked items")
        self.toolbar.addWidget(self.btn_delete)

        self.thread_pool = QThreadPool()
        model = TreeViewStandardItemModel(["Name"])
        self.treeview.setModel(model)
        self.populate_treeview()

    def check_all_items(self) -> None:
        """
        Check all items.
        """
        try:
            model: TreeViewStandardItemModel = self.treeview.model()
            model.check_all_items()
        except Exception as e:
            print(e)

    def uncheck_all_items(self) -> None:
        """
        Uncheck all items.
        """
        try:
            model: TreeViewStandardItemModel = self.treeview.model()
            model.uncheck_all_items()
        except Exception as e:
            print(e)

    def edit_selected_item(self) -> None:
        """
        Edit the selected item.
        :return:
        """
        model = self.treeview.model()
        selected_indexes = self.treeview.selectionModel().selectedIndexes()
        if not selected_indexes:
            GuiUtils.show_error("No item selected.")
            return

        selected_index = selected_indexes[0]
        treeview_item = model.itemFromIndex(selected_index)

        item_model: MenuItem = treeview_item.tag
        edit_item_form = MenuItemForm()
        edit_item_form.setFixedSize(600, 500)
        edit_item_form.form_data = {
            "name": item_model.name,
            "icon": item_model.icon,
            "models": item_model.models,
            "capture_mode": item_model.capture_mode,
            "system_instruction": item_model.system_instruction,
            "guidance_prompt": item_model.guidance_prompt
        }
        dialog_result = edit_item_form.exec()
        if dialog_result == QDialog.DialogCode.Accepted:
            item_model.name = edit_item_form.name
            item_model.icon = edit_item_form.icon
            item_model.models = edit_item_form.models
            item_model.capture_mode = edit_item_form.capture_mode
            item_model.system_instruction = edit_item_form.system_instruction
            item_model.guidance_prompt = edit_item_form.guidance_prompt
            self.save_or_update_item(item_model)

    def save_or_update_item(self, item):
        """
        Save or update an item.
        :param item:
        :param item_id:
        :return:
        """
        work = Worker(self.do_save_or_update_item, item)
        work.signals.result.connect(self.on_save_or_update_done)
        self.thread_pool.start(work)

    @work_exception
    def do_save_or_update_item(self, item):
        """
        Do the work -> save or update the item.
        :param item:
        :return:
        """
        if item.icon:
            icon_in_path = Path(item.icon)
            quack2tex_home = LibUtils.get_lib_home()
            icons_folder = quack2tex_home / "icons"
            icons_folder.mkdir(exist_ok=True, parents=True)
            image: PILImage = Image.open(str(icon_in_path))
            image.thumbnail((64, 64))
            icon_out_path = icons_folder / f"{icon_in_path.name}"
            image.save(str(icon_out_path))
            item.icon = str(icon_out_path)

        try:
            with get_db_session() as session:
                if item.id:
                    print("Updating item")
                    new_item = MenuItemRepository.update_item(session, item)
                else:
                    new_item = MenuItemRepository.add_item(session, item)
            print(f"Item saved: {new_item}")
            return new_item
        except Exception as e:
            if icon_out_path.exists():
                icon_out_path.unlink()
            raise e

    def on_save_or_update_done(self, result) -> None:
        """
        Handle the completion of save or update operation.
        :param result:
        :return:
        """
        new_item, error = result
        if error:
            GuiUtils.show_error(str(error))
            return
        GuiUtils.show_info("Item updated successfully.")
        self.on_settings_changed.emit()
        self.populate_treeview()

    def delete_selected_items(self) -> None:
        """
        Delete selected items.
        """
        model: TreeViewStandardItemModel = self.treeview.model()
        checked_items = model.find_checked_nodes_recursive()
        if not checked_items:
            GuiUtils.show_error("No items selected.")
            return
        self.populate_treeview()

        ids = [item.tag.id for item in checked_items]
        work = Worker(self.do_delete_items, ids)
        work.signals.result.connect(self.on_delete_done)
        self.thread_pool.start(work)

    def do_delete_items(self, ids):
        """
        Do the work -> delete the items.
        :param ids:
        :return:
        """
        with get_db_session() as session:
            MenuItemRepository.delete_items(session, ids)

    def on_delete_done(self) -> None:
        """
        Handle the completion of delete operation.
        :return:
        """
        GuiUtils.show_info("Items deleted successfully.")
        self.on_settings_changed.emit()
        model: TreeViewStandardItemModel = self.treeview.model()
        model.remove_checked_items()
        model.layoutChanged.emit()
        self.populate_treeview()

    def add_new_item(self) -> None:
        """
        Add a new item.
        """
        selected_indexes: List[QModelIndex] = self.treeview.selectionModel().selectedIndexes()
        parent_item: Optional[QStandardItem] = None
        if selected_indexes:
            selected_index: QModelIndex = selected_indexes[0]
            model: TreeViewStandardItemModel = self.treeview.model()
            parent_item = model.itemFromIndex(selected_index)

        new_item_form = MenuItemForm()
        new_item_form.setFixedSize(600, 500)
        dialog_result = new_item_form.exec()
        if dialog_result == QDialog.DialogCode.Accepted:
            new_item = MenuItem(
                name=new_item_form.name,
                icon=new_item_form.icon,
                models=new_item_form.models,
                capture_mode=new_item_form.capture_mode,
                system_instruction=new_item_form.system_instruction,
                guidance_prompt=new_item_form.guidance_prompt,
                parent_id=parent_item.tag.id if parent_item else None
            )
            self.save_or_update_item(new_item)

    def populate_treeview(self) -> None:
        """
        Populate the treeview with data from the database.
        """
        work = Worker(self.do_fetch_tree_data)
        work.signals.result.connect(self.on_tree_data_fetched)
        self.thread_pool.start(work)

    def do_fetch_tree_data(self) -> List[MenuItem]:
        """
        Do the work -> fetch the data from the database.
        :return:
        """
        with get_db_session() as session:
            return MenuItemRepository.build_tree(session)

    def on_tree_data_fetched(self, result: List[MenuItem]) -> None:
        """
        Handle the completion of tree data fetch operation.
        :param result:
        :return:
        """
        self.treeview.model().reset_model()

        # Main iteration to populate the root items
        for tree_item_data in result:
            treeview_item = QStandardItem(tree_item_data.name)
            treeview_item.setCheckable(True)
            treeview_item_icon = QIcon(":/icons/gears.png")
            if tree_item_data.icon:
                treeview_item_icon = QIcon(tree_item_data.icon)
            treeview_item.setIcon(treeview_item_icon)
            treeview_item.tag = tree_item_data
            self.treeview.model().appendRow(treeview_item)
            self.populate_children(tree_item_data, treeview_item)

    def populate_children(self,tree_item_data: MenuItem, parent_item: QStandardItem) -> None:
        """
        Populate the children of the tree item in an optimized, non-recursive manner.
        :param tree_item_data: The MenuItem containing child items.
        :param parent_item: The QStandardItem that will be the parent of child items.
        :return: None
        """
        items_to_add = []

        # Pre-create child items and prepare them for appending
        for child_item_data in tree_item_data.children:
            child_item = QStandardItem(child_item_data.name)
            child_item.setCheckable(True)
            child_item.setIcon(QIcon(child_item_data.icon))
            child_item.tag = child_item_data
            items_to_add.append(child_item)

        # Append all child items at once
        if items_to_add:
            parent_item.appendRows(items_to_add)

        # Now populate the children for all the new child items in a non-recursive manner
        for child_item, child_data in zip(items_to_add, tree_item_data.children):
            self.populate_children(child_data, child_item)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    win = SettingsWindow()
    win.show()
    sys.exit(app.exec())


from enum import Enum

from quack2tex.pyqt import (
    Qt, QTimer, QPropertyAnimation, QEasingCurve,QFont,QWidget, QLabel, QVBoxLayout, QGraphicsOpacityEffect
)

class ToastType(Enum):
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class Toast(QWidget):
    """
    A toast message that displays a message for a short period of time
    """

    def __init__(self, message, message_type=ToastType):
        super().__init__()
        self.setWindowTitle("Toast")
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        font = QFont()
        font.setPointSize(14)
        font.setBold(True)

        styles = {
            ToastType.INFO: "background-color: #1ffc64; color: black;",
            ToastType.WARNING: "background-color: #ffe146; color: black;",
            ToastType.ERROR: "background-color: #e71313; color: white;",
        }
        if message_type not in styles:
            message_type = ToastType.INFO
        self.setStyleSheet(styles[message_type])
        self.label = QLabel(message, self)
        self.label.setContentsMargins(10, 10, 10, 0)
        self.label.setFont(font)
        self.label.setAlignment(
            Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignHCenter
        )
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)  # Remove layout margins
        self.layout.addWidget(self.label)
        self.adjustSize()

        # Set up opacity effect and animation
        self.effect = QGraphicsOpacityEffect(self)
        self.setGraphicsEffect(self.effect)
        self.animation = QPropertyAnimation(self.effect, b"opacity")
        self.animation.setDuration(500)  # 500 milliseconds
        self.animation.setEasingCurve(QEasingCurve.Type.InOutQuad)

    def showEvent(self, event):
        """
        Show the toast message with fade-in effect and auto close after 5 seconds
        """
        self.fade_in()
        QTimer.singleShot(3000, self.fade_out)  # Show for 5 seconds, then fade out

    def fade_in(self):
        """
        Fade in the toast message
        """
        self.animation.setStartValue(0)
        self.animation.setEndValue(1)
        self.animation.start()

    def fade_out(self):
        """
        Fade out the toast message
        """
        self.animation.setStartValue(1)
        self.animation.setEndValue(0)
        self.animation.finished.connect(self.close)  # Close widget after fade out
        self.animation.start()



from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *
from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot
from PyQt6.QtCore import pyqtProperty as Property
from PyQt6.QtWebChannel  import QWebChannel
from PyQt6.QtWebEngineWidgets import QWebEngineView
from PyQt6.QtWebEngineCore import QWebEngineSettings, QWebEnginePage





from __future__ import annotations

import os
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, List

from sqlalchemy import ForeignKey, event
from sqlalchemy.orm import DeclarativeBase, relationship
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import mapped_column

from .db.registry import mapper_registry as reg


class Base(MappedAsDataclass, DeclarativeBase):
    registry = reg


class MenuItem(Base):
    """
    Block model.
    """

    __tablename__ = "items"

    id: Mapped[int] = mapped_column(init=False, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(nullable=False)
    icon: Mapped[str] = mapped_column(nullable=False)
    system_instruction: Mapped[str] = mapped_column(nullable=True, default=None)
    guidance_prompt: Mapped[str] = mapped_column(nullable=True, default=None)
    models: Mapped[str] = mapped_column(nullable=True, default=None)
    capture_mode: Mapped[str] = mapped_column(nullable=True, default=None)
    parent_id: Mapped[int] = mapped_column(ForeignKey("items.id"), nullable=True, default=None)

    # Define relationships between parent and children
    if TYPE_CHECKING:
        parent: "MenuItem"
        children: List["MenuItem"] = []
    else:
        # Avoid using the same name 'parent' in backref
        parent = relationship("MenuItem", remote_side=[id])
        children = relationship('MenuItem', remote_side=[parent_id], uselist=True, back_populates='parent', cascade="all, delete-orphan")

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.id}, {self.name})>"

@event.listens_for(MenuItem, 'after_delete')
def receive_after_delete(mapper, connection, target):
    """
    Event listener for after delete
    :param mapper:
    :param connection:
    :param target:
    :return:
    """
    # remove logo file
    print(f"Deleting logo file: {target.icon}")
    if not target.icon:
        return
    logo_path = Path(target.icon)
    if logo_path.exists():
        os.remove(logo_path)


import os

import typer
from dotenv import load_dotenv, find_dotenv

import quack2tex

app = typer.Typer(invoke_without_command=True)


@app.command()
def start(
    gemini_api_key: str = typer.Option(None, envvar="GEMINI_API_KEY", help="Google Gemini API key"),
    openai_api_key: str = typer.Option(None, envvar="OPENAI_API_KEY", help="OpenAI API key"),
    anthropic_api_key: str = typer.Option(None, envvar="ANTHROPIC_API_KEY", help="Anthropic API key"),
    groq_api_key: str = typer.Option(None, envvar="GROQ_API_KEY", help="Groq API key"),
):
    """
    Start the Quack2Tex application with optional LLM API keys.
    You can provide keys via command-line or environment variables.
    """
    # Set API keys in environment if provided
    api_keys = {
        "GEMINI_API_KEY": gemini_api_key,
        "OPENAI_API_KEY": openai_api_key,
        "ANTHROPIC_API_KEY": anthropic_api_key,
        "GROQ_API_KEY": groq_api_key,
    }

    for key, value in api_keys.items():
        if value:
            os.environ[key] = value

    # Start the main app
    quack2tex.run_app()


def run():
    """
    Entry point: Load environment variables and invoke CLI app.
    """
    load_dotenv(find_dotenv())
    app()


if __name__ == "__main__":
    run()



import sys
import traceback
from functools import wraps

from quack2tex.utils import GuiUtils


def gui_exception(func):
    """
    A decorator that wraps the passed in function and logs
    exceptions should one occur
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as ex:
            print(f"Exception in {func.__name__}: {ex}")
            exc_type, exc_value, exc_traceback = sys.exc_info()
            GuiUtils.show_error_message(
                f"{str(traceback.format_exception_only( exc_type, exc_value )[0])}"
            )

    return wrapper



import io

from quack2tex.pyqt import (
    Qt, QSize, QMovie, QCloseEvent, QShowEvent, QShortcut, QLabel, QWidget, QVBoxLayout, QResource
)
from PIL import Image
from quack2tex.utils import GuiUtils


class LoadingIndicator(QWidget):
    """
    A loading indicator that displays a GIF animation
    """

    def __init__(self, gif_path, gif_size=QSize(100, 100), parent=None):
        super().__init__(parent)

        self.setWindowTitle("Loading Indicator")
        # self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        resource = QResource(gif_path)
        image_data = resource.data()
        image_bytes = io.BytesIO(image_data)
        gif_img = Image.open(image_bytes)
        gif_full_sz = QSize(gif_img.size[0], gif_img.size[1])
        del gif_img
        target_sz = GuiUtils.calculate_new_size(gif_full_sz, gif_size)
        self.label = QLabel(self)
        self.label.setStyleSheet("background-color: rgba(255, 255, 255, 0)")
        self.label.setFixedSize(target_sz)
        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.movie = QMovie(gif_path)
        self.movie.setScaledSize(target_sz)
        self.label.setMovie(self.movie)
        # Create a layout and add the QLabel to it
        layout = QVBoxLayout(self)
        layout.addWidget(self.label)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.adjustSize()

        QShortcut("Ctrl+C", self, self.close)

    def start(self):
        """
        Start the loading animation
        """
        if self.movie.state() == QMovie.MovieState.NotRunning:
            self.movie.start()

    def stop(self):
        """
        Stop the loading animation
        """
        if self.movie.state() == QMovie.MovieState.Running:
            self.movie.stop()

    def showEvent(self, e: QShowEvent):
        """
        :param e: QShowEvent
        """
        self.start()
        super().showEvent(e)

    def keyPressEvent(self, event):
        """
        :param event: QKeyEvent
        """
        if event.key() == Qt.Key.Key_Escape:
            self.close()
        else:
            super().keyPressEvent(event)

    def closeEvent(self, e: QCloseEvent):
        """
        :param e: QCloseEvent
        """

        self.stop()

        super().closeEvent(e)



from PIL.Image import Image as PILImage
import base64
from io import BytesIO


class ImageUtils:
    """
    Utility functions for images.
    """

    @staticmethod
    def image_to_base64(image: PILImage) -> str:
        """
        Convert an image to a base64 string.
        :param image:
        :return:
        """
        buffered = BytesIO()
        image.save(
            buffered, format=image.format or "PNG"
        )  # Default to 'PNG' if format is None
        return base64.b64encode(buffered.getvalue()).decode("utf-8")

    @staticmethod
    def image_to_base64_url(image: PILImage) -> str:
        """Convert an image to a base64 data URL."""
        return f"data:image/{image.format.lower()};base64,{ImageUtils.image_to_base64(image)}"



from .worker import Worker
from .gui_utils import GuiUtils
from .image_utils import ImageUtils
from .lib_utils import LibUtils
from .singleton import Singleton
from .work_exception import work_exception
from .treeview_standard_model import TreeViewStandardItemModel



from quack2tex.pyqt import (
    Property,
QFileDialog, QHBoxLayout, QWidget, QPushButton
)
from quack2tex.widgets import ClickableLineEdit


class FileUploader(QWidget):
    """
    A custom widget that allows the user to upload a file
    """

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent)

        # Create a horizontal layout for the uploader widget
        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)

        # Create the text field to display the file path
        self.file_path_edit = ClickableLineEdit(self)
        self.file_path_edit.setFixedHeight(30)
        self.file_path_edit.doubleClicked.connect(self.open_file_dialog)
        self.file_path_edit.setPlaceholderText("No file selected")

        # Create the upload button
        self.upload_button = QPushButton("Browse...", self)
        self.upload_button.clicked.connect(self.open_file_dialog)

        # Add the text field and button to the layout
        self.layout.addWidget(self.file_path_edit)
        self.layout.addWidget(self.upload_button)

        # Set the layout on this custom widget
        self.setLayout(self.layout)
        self.file_path_edit.setReadOnly(True)

        # Set the file filter
        file_filter = kwargs.get("file_filter", None)
        self.file_filter = file_filter or "All Files (*)"


    def get_file_path(self):
        """
        Get the file path
        :return: the file path
        """
        return self.file_path_edit.text()


    def set_file_path(self, file_path):
        """
        Set the file path
        :param file_path:
        :return:
        """
        self.file_path_edit.setText(file_path)

    file_path = Property(str, get_file_path, set_file_path)


    def open_file_dialog(self):
        """
        Open a file dialog to select a file
        :return:
        """
        file_path, _ = QFileDialog.getOpenFileName(self, "Select File", "", self.file_filter)
        if file_path:
            self.file_path_edit.setText(file_path)

    def mouseDoubleClickEvent(self, event):
        """
        Triggered when the user double-clicks the widget
        :param event:
        :return:
        """
        self.open_file_dialog()
        super().mouseDoubleClickEvent(event)


import inspect
import sys
import typing

from pydantic import BaseModel

from quack2tex.pyqt import (
    QApplication, Qt, QIcon, QFile, QPalette, QColor, QCursor, QFontDatabase,QIODevice
)
from modihub.llm import LLM
from .resources import * # noqa
from .utils import GuiUtils
from .windows import MainWindow


class Quack2TexWrappedFunctionResult(BaseModel):
      result: typing.Any
      latex: str 

def apply_theme(app: QApplication) -> None:
    """
    Apply the theme to the application
    :param app:
    :return:
    """
    app.setStyle("Fusion")
    QFontDatabase.addApplicationFont(":/fonts/Roboto/Roboto-Regular.ttf")
    app.setWindowIcon(QIcon(":icons/rubber-duck.png"))
    file = QFile(":/styles/style.qss")
    file.open(QIODevice.OpenModeFlag.ReadOnly | QIODevice.OpenModeFlag.Text)
    style_sheet = file.readAll().data().decode("utf-8")
    app.setStyleSheet(style_sheet)
    palette = QPalette()
    palette.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.WindowText, QColor("white"))
    palette.setColor(QPalette.ColorRole.Base, QColor(25, 25, 25))
    palette.setColor(QPalette.ColorRole.AlternateBase, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ToolTipBase, QColor("white"))
    palette.setColor(QPalette.ColorRole.ToolTipText, QColor("white"))
    palette.setColor(QPalette.ColorRole.Text, QColor("white"))
    palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ButtonText, QColor("white"))
    palette.setColor(QPalette.ColorRole.BrightText, QColor("red"))
    palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.HighlightedText, QColor("black"))
    QApplication.setPalette(palette)


# create quack2tex decorator
def latify(model: str = None):
    """
    Decorator to convert a function to LaTeX using the LLM model
    :param kwargs:
    :return:
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            llm = LLM.create(model or "models/gemini-1.5-flash-latest")
            func_code = inspect.getsource(func)
            prompt = kwargs.get("prompt", "Generate a LaTeX representation of the following function in markdown format:")
            response = llm.ask(prompt + "\n\n" + func_code)
            result = func(*args, **kwargs)
            return Quack2TexWrappedFunctionResult(result=result, latex=response)
        return wrapper
    return decorator

def run_app() -> None:
    """
    Run the application.
    :return:
    """

    app = QApplication(sys.argv)
    apply_theme(app)
    app.setOverrideCursor(QCursor(Qt.CursorShape.PointingHandCursor))
    window = MainWindow()
    window.show()
    sys.exit(app.exec())



import weakref
from quack2tex.pyqt import Signal, QWidget
from .floating_menu_item import FloatingMenuItem


class FloatingMenu(QWidget):
    """
    A floating menu widget that displays a circular menu
    """
    item_clicked = Signal(dict)

    def __init__(self, root_item: FloatingMenuItem = None, parent=None):
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.root_item = weakref.ref(root_item) if root_item else None  # Safely create weakref if root_item is not None

    def set_root(self, root_item: FloatingMenuItem):
        """
        Set the root item of the menu.
        :param root_item:
        :return:
        """
        self.root_item = weakref.ref(root_item) if root_item else None  # Safely create weakref if root_item is not None
        if self.root_item:
            self.root_item().setParent(self)  # Dereference before use

    def handle_item_click(self):
        """
        Handle the item click event.
        :return:
        """
        try:
            if current_item := self.sender():
                if current_item is not None:  # Check if the object is still valid
                    current_item.toggle()
                    data = current_item.data
                    if data:
                        self.item_clicked.emit(data)
                        self.toggle_item(current_item)

                    if current_item == self.root_item() and self.root_item() and self.root_item().is_expanded():
                        self.root_item().toggle()
        except Exception as e:
            print(e)

    def toggle_item(self, item):
        """
        Recursively toggle the root item.
        :param item:
        :return:
        """
        if item.root:
            if item.root is not None:  # Check if the root item still exists
                item.root.toggle()
                self.toggle_item(item.root)

    def draw_item_children(self, parent_item: FloatingMenuItem, center):
        """
        Draw the children of a given item.
        :param parent_item:
        :param center:
        :return:
        """
        try:
            for item in parent_item.children:
                if item is not None:  # Check if item is valid before using
                    item.doubleClicked.connect(self.handle_item_click)
                    item.setParent(self)
                    item.root = parent_item
                    item.move(center.x() - item.width() // 2, center.y() - item.height() // 2)
                    item.hide()
                    if item.children:
                        self.draw_item_children(item, center)
        except Exception as e:
            print(e)

    def draw_menu(self):
        """
        Draw the menu
        :return:
        """
        try:
            if self.root_item() is not None:  # Check if root item is valid before using
                center = self.geometry().center()
                self.root_item().move(center.x() - self.root_item().width() // 2, center.y() - self.root_item().height() // 2)
                self.root_item().doubleClicked.connect(self.handle_item_click)
                self.root_item().show()
                if self.root_item().children:
                    self.draw_item_children(self.root_item(), center)
            else:
                print("Root item is not set.")
        except Exception as e:
            print(e)


import contextlib
from typing import Any, AsyncIterator

from sqlalchemy import event, NullPool, create_engine
from sqlalchemy.ext.asyncio import (
    AsyncConnection,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import sessionmaker
from .registry import mapper_registry
from quack2tex.utils import Singleton


class SessionManager(metaclass=Singleton):
    """
    A context manager for creating database sessions
    """

    def __init__(
        self,
        url: str,
        enable_foreign_keys: bool = True,
        async_mode: bool = False,
        session_kwargs: dict | None = None,
        engine_kwargs: dict | None = None,
    ):
        """
        Initialize the SessionManager.

        :param url: Database URL
        :param enable_foreign_keys: Enable foreign key constraints
        :param async_mode: Enable asynchronous mode
        :param session_kwargs: Additional session keyword arguments
        :param engine_kwargs: Additional engine keyword arguments
        """
        self._async_mode = async_mode
        engine_kwargs = engine_kwargs or {}
        session_kwargs = session_kwargs or {
            "autocommit": False,
            "autoflush": False,
            "expire_on_commit": False,
        }

        # Create the appropriate engine based on async_mode.
        if self._async_mode:
            self._engine = create_async_engine(url, **engine_kwargs)
            SessionFactory = async_sessionmaker
        else:
            self._engine = create_engine(url, **engine_kwargs)
            SessionFactory = sessionmaker

        # Apply common session settings
        self._session_maker = SessionFactory(**session_kwargs, bind=self._engine)
        self._session = None
        if enable_foreign_keys:
            engine = self._engine if not self._async_mode else self._engine.sync_engine

            @event.listens_for(
                engine,
                "connect",
            )
            def _fk_pragma_on_connect(dbapi_con, con_record):
                dbapi_con.execute("pragma foreign_keys=ON")

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Create a new SessionManager instance and initialize it.

        :return: Initialized SessionManager instance
        """
        session_manager = cls(*args, **kwargs)
        session_manager.init()
        return session_manager

    @classmethod
    async def async_create(cls, *args, **kwargs):
        """
        Create a new SessionManager instance and initialize it asynchronously.

        :return: Initialized SessionManager instance
        """
        session_manager = cls(*args, **kwargs)
        await session_manager.async_init()
        return session_manager

    def __enter__(self):
        """
        Enter the context manager, creating a new session.

        :return: Database session
        """
        assert not self._async_mode, "Asynchronous mode is enabled"
        self._session = self._session_maker()
        return self._session

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the context manager, closing the session.

        :param exc_type: Exception type
        :param exc_val: Exception value
        :param exc_tb: Exception traceback
        """
        if exc_val:
            self._session.rollback()
        self._session.close()

    async def __aenter__(self):
        """
        Enter the async context manager, creating a new async session.

        :return: Async database session
        """
        assert self._async_mode, "Asynchronous mode is not enabled"
        self._session = self._session_maker()
        return self._session

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the async context manager, closing the async session.

        :param exc_type: Exception type
        :param exc_val: Exception value
        :param exc_tb: Exception traceback
        """
        if exc_val:
            await self._session.rollback()
        await self._session.close()

    @contextlib.asynccontextmanager
    async def async_session(self, *args, **kwargs) -> AsyncIterator[AsyncSession]:
        """
        Create an async session.

        :yield: Async database session
        """
        assert self._async_mode, "Asynchronous mode is not enabled"
        async with self._session_maker(*args, **kwargs) as session:
            try:
                yield session
            except Exception:
                await session.rollback()
                raise

    @contextlib.contextmanager
    def session(self, *args, **kwargs) -> Any:
        """
        Create a session.

        :yield: Database session
        """
        assert not self._async_mode, "Asynchronous mode is enabled"
        session = self._session_maker(*args, **kwargs)
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    async def async_init(self, drop_all: bool = False):
        """
        Initialize the database asynchronously.

        :param drop_all: Drop all tables before creating them
        :return: Self
        """
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")
        async with self._engine.begin() as conn:
            if drop_all:
                await conn.run_sync(mapper_registry.metadata.drop_all)
            await conn.run_sync(mapper_registry.metadata.create_all)
        return self

    def init(self, drop_all: bool = False):
        """
        Initialize the database.

        :param drop_all: Drop all tables before creating them
        :return: Self
        """
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")
        if drop_all:
            mapper_registry.metadata.drop_all(self._engine)
        mapper_registry.metadata.create_all(self._engine)
        return self

    async def async_close(self):
        """
        Close the database asynchronously.
        """
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")
        await self._engine.dispose()
        self._engine = None
        self._session_maker = None

    def close(self):
        """
        Close the database.
        """
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")
        self._engine.dispose()
        self._engine = None
        self._session_maker = None

    @contextlib.asynccontextmanager
    async def async_connect(self) -> AsyncIterator[AsyncConnection]:
        """
        Connect to the database asynchronously.

        :yield: Async database connection
        """
        assert self._async_mode, "Asynchronous mode is not enabled"
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")

        async with self._engine.begin() as connection:
            try:
                yield connection
            except Exception:
                await connection.rollback()
                raise

    @contextlib.contextmanager
    def connect(self) -> Any:
        """
        Connect to the database.

        :yield: Database connection
        """
        assert not self._async_mode, "Asynchronous mode is enabled"
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")

        connection = self._engine.connect()
        try:
            yield connection
        except Exception:
            connection.rollback()
            raise
        finally:
            connection.close()




import typing
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

from PIL.Image import Image as PILImage
from tqdm import tqdm

from modihub.llm import LLM
from quack2tex.pyqt import (
    Qt,
    QThreadPool,
    QMainWindow,
)
from quack2tex.utils import GuiUtils, Worker, work_exception
from quack2tex.widgets import DuckMenu
from .ouput_dialog import OutputDialog
from .screen_capture import ScreenCaptureWindow
from .settings_window import SettingsWindow


class MainWindow(QMainWindow):
    """
    Main application window.
    """

    def __init__(self):
        super().__init__()

        # Window settings
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        self.menu = DuckMenu()
        self.menu.setFixedSize(400, 400)
        self.menu.build_menu()
        self.menu.item_clicked.connect(self.handle_menu_item_click)
        self.setCentralWidget(self.menu)
        self.threadpool = QThreadPool()

        # drag and drop variables
        self.is_moving = False
        self.offset = None

    def handle_menu_item_click(self, data):
        """
        Handle the menu item click event.
        :param data:
        :return:
        """
        menu_item_action = data.get("action", None)
        menu_item_data = data.get("tag", None)
        if menu_item_action == "exit":
            self.close()
        elif menu_item_action == "settings":
            w = SettingsWindow()
            w.on_settings_changed.connect(self.menu.build_menu)
            w.exec()
        elif menu_item_data:
            capture_mode = menu_item_data.capture_mode
            prompt_data = {
                "system_instruction": menu_item_data.system_instruction,
                "guidance_prompt": menu_item_data.guidance_prompt,
                "models": menu_item_data.models,
                "capture_mode": capture_mode
            }
            if capture_mode == "screen":
                self.start_screen_capture(prompt_data)
            elif capture_mode == "clipboard":
                self.start_clipboard_text_capture(prompt_data)
            else:
                self.make_prompt_request(prompt_data, prompt_input="")

    def pick_screen_region(self):
        """
        Pick the screen region
        :return:
        """
        screen_capture = ScreenCaptureWindow()
        monitor_geometry = GuiUtils.get_current_monitor_geometry(self)
        screen_capture.setGeometry(monitor_geometry)
        screen_capture.exec()
        return screen_capture.selected_region



    def start_screen_capture(self, prompt_data):
        """
        Start the screen capture process
        :param prompt_data:
        :return:
        """
        monitor_index = GuiUtils.get_current_monitor_index(self)
        screen_region = self.pick_screen_region()
        if screen_region:
            @work_exception
            def do_work():
                """
                Perform the screen capture
                :return:
                """
                return GuiUtils.get_screen_capture_image(screen_region, monitor_index)
            def done(result):
                """
                Handle the completion of the screen capture
                :param result:
                :return:
                """
                screen_capture, error = result
                if error:
                    GuiUtils.show_error(str(error))
                    return
                self.make_prompt_request(prompt_data, prompt_input=screen_capture)
            worker = Worker(do_work)
            worker.signals.result.connect(done)
            self.threadpool.start(worker)

    def start_clipboard_text_capture(self, prompt_data):
        """
        Start the clipboard text capture process
        :param prompt_data:
        :return:
        """
        clipboard_text = GuiUtils.get_clipboard_text()
        if clipboard_text is None:
            GuiUtils.show_error("No text copied to clipboard.")
            return
        self.make_prompt_request(prompt_data, prompt_input=clipboard_text)

    @work_exception
    def make_prompt_request_do_work(self, prompt_data: dict, prompt_input: typing.Union[str,PILImage]):
        """
        Start the prompt data capture process
        :param prompt_data:
        :param prompt_input:
        :param kwargs:
        :return:
        """
        return self.process_prompt_request(prompt_data, prompt_input)

    def make_prompt_request_done(self, result):
        """
        Handle the completion of the screen capture and description generation
        :param result:
        :return:
        """
        predictions, error = result
        if error:
            GuiUtils.show_error(str(error))
            return
        self.menu.loading_indicator.close()
        self.create_output_dialog(predictions)

    def make_prompt_request(self, prompt_data: dict, prompt_input: typing.Union[str,PILImage]):
        """
        Start the prompt data capture process
        :param prompt_data:
        :param prompt_input:
        :param kwargs:
        :return:
        """
        self.menu.loading_indicator.show()

        worker = Worker(self.make_prompt_request_do_work, prompt_data, prompt_input)
        worker.signals.result.connect(self.make_prompt_request_done)
        self.threadpool.start(worker)

    def create_output_dialog(self, predictions: dict):
        """
        Create an output window
        :param text:
        :return:
        """
        dialog = OutputDialog(predictions, parent=self)
        dialog.setWindowTitle("Output")
        dialog.adjustSize()
        GuiUtils.move_window_to_center(dialog)
        dialog.activateWindow()
        dialog.exec()

    @staticmethod
    def call_llm(model, system_instruction, multimodal_prompt):
        """
        Standalone function to call the language model
        :param model:
        :param system_instruction:
        :param multimodal_prompt:
        :return:
        """
        llm = LLM.create(model, system_instruction=system_instruction)
        return llm(multimodal_prompt)


    def process_prompt_request(self, prompt_data: dict, prompt_input:  typing.Union[str,PILImage]) -> dict:
        """
        Call the language model
        :param prompt_data:
        :param prompt_input:
        :return:
        """
        models = prompt_data.get("models")
        system_instruction = prompt_data.get("system_instruction")
        guidance_prompt = prompt_data.get("guidance_prompt")
        multimodal_prompt = [guidance_prompt, prompt_input]

        models  = models.split(",") if models else []
        results = {}
        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(self.call_llm, model, system_instruction, multimodal_prompt): model
                for model in models
            }
            for future in tqdm(as_completed(futures), total=len(futures)):
                model_name = futures[future]
                try:
                    results[model_name] = future.result()
                except Exception as e:
                    results[model_name] = f"Error by running inference on model {model_name}: {e}"
        return results



    def mousePressEvent(self, event):
        """
        Triggered when the user presses the mouse button.
        :param event:
        :return:
        """
        if event.button() == Qt.MouseButton.LeftButton:
            self.is_moving = True
            self.offset = event.globalPosition().toPoint() - self.frameGeometry().topLeft()

    def mouseMoveEvent(self, event):
        """
        Triggered when the user moves the mouse.
        :param event:
        :return:
        """
        if self.is_moving:
            new_pos = event.globalPosition().toPoint() - self.offset
            self.move(new_pos)

    def mouseReleaseEvent(self, event):
        """
        Triggered when the user releases the mouse button.
        :param event:
        :return:
        """
        if event.button() == Qt.MouseButton.LeftButton:
            self.is_moving = False



import typing

from quack2tex.pyqt import (
    Signal, Property, QWidget, QButtonGroup, QRadioButton, QHBoxLayout
)


class RadioButtonGroup(QWidget):
    """
    A group of radio buttons
    """

    on_radio_button_clicked = Signal(object)


    def __init__(self, parent=None):
        super(RadioButtonGroup, self).__init__(parent)
        # Create the layout

        # Create the layout
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        # Create a button group
        self.radio_group = QButtonGroup(self)
        self.radio_group.buttonClicked.connect(self.handler_on_radio_button_clicked)

        # Set the layout of the window
        self.setLayout(layout)

    @Property(str)
    def selected_option(self):
        """
        Get the selected option
        :return: the selected option
        """
        selected_button = self.radio_group.checkedButton()
        return selected_button.property("value") if selected_button else "None"

    def set_select_option(self, value):
        """
        Select an option
        :param value: the value to select
        :return: None
        """
        for button in self.radio_group.buttons():
            if button.property("value") == value:
                print("Setting button")
                button.setChecked(True)
                break

    def handler_on_radio_button_clicked(self, button):
        """
        Handler for when a radio button is clicked
        :param button: the button that was clicked
        :return: None
        """
        self.on_radio_button_clicked.emit(button.property("value"))

    def add_option(self, text_value : str,checked:bool = False, value: typing.Any =None):
        """
        Add a new radio button to the group
        :param text_value: the text value to display
        :param checked: whether the radio button should be checked
        :param value: the data to associate with the radio button
        :return: None
        """
        radio_button = QRadioButton(text_value)
        radio_button.setChecked(checked)
        radio_button.setProperty("value", value)
        self.radio_group.addButton(radio_button)
        self.layout().addWidget(radio_button)




from threading import Lock


class Singleton(type):
    """
    This is a thread-safe implementation of Singleton.
    """

    _instances: dict = {}

    _lock: Lock = Lock()
    """
    We now have a lock object that will be used to synchronize threads during
    first access to the Singleton.
    """

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        # Now, imagine that the program has just been launched. Since there's no
        # Singleton instance yet, multiple threads can simultaneously pass the
        # previous conditional and reach this point almost at the same time. The
        # first of them will acquire lock and will proceed further, while the
        # rest will wait here.
        with cls._lock:
            # The first thread to acquire the lock, reaches this conditional,
            # goes inside and creates the Singleton instance. Once it leaves the
            # lock block, a thread that might have been waiting for the lock
            # release may then enter this section. But since the Singleton field
            # is already initialized, the thread won't create a new object.
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]



from .image_button import ImageButton
from .floating_menu import FloatingMenu, FloatingMenuItem
from .toast import Toast, ToastType
from .loading_indicator import LoadingIndicator
from .mardown_viewer import MarkdownViewer
from .duck_menu import DuckMenu
from .clickable_line_edit import ClickableLineEdit
from .file_uploader import FileUploader
from .radio_button_group import RadioButtonGroup
from .model_picker import ModelPicker



from functools import wraps


def work_exception(function):
    """
    A decorator that wraps the passed in function and logs
    exceptions should one occur
    """

    @wraps(function)
    def wrapper(*args, **kwargs):
        try:
            if args or kwargs:
                output = function(*args, **kwargs)
            else:
                output = function()
            return output, None
        except Exception as ex:
            return None, ex

    return wrapper



from quack2tex.pyqt import (
    Signal,
    QLineEdit
)

class ClickableLineEdit(QLineEdit):
    """
    A custom QLabel that emits a doubleClicked signal when double clicked
    """

    doubleClicked = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

    def mouseDoubleClickEvent(self, event):
        """
        Emit the doubleClicked signal on double click
        :param event:
        :return:
        """
        self.doubleClicked.emit()
        super().mouseDoubleClickEvent(event)


from quack2tex.pyqt import (
    Qt, Property, QThreadPool, Signal, QIcon, QListWidget, QListWidgetItem,
    QWidget, QVBoxLayout, QLabel, QMenu
)
from quack2tex.resources import *  # noqa: F401


class ModelPicker(QWidget):
    """
    A widget that allows the user to select a model.
    """
    on_loaded_models = Signal()

    def __init__(self, parent=None):
        super(ModelPicker, self).__init__(parent)

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.thread_pool = QThreadPool()

        self.list_widget = QListWidget()
        self.list_widget.setSelectionMode(QListWidget.SelectionMode.MultiSelection)
        self.list_widget.itemSelectionChanged.connect(self.on_selection_changed)

        self.selected_models_label = QLabel()
        self.layout.addWidget(self.list_widget)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.open_context_menu)

    def set_data(self, models):
        """
        Load the models into the list widget.
        """
        icons = {
            "ollama": QIcon(":/icons/ollama.png"),
            "google": QIcon(":/icons/google.png"),
            "openai": QIcon(":/icons/openai.png"),
            "anthropic": QIcon(":/icons/anthropic.png"),
            "groq": QIcon(":/icons/groq.png"),
        }
        for model in models:
            model_item = QListWidgetItem(model.display_name)
            model_item.setData(Qt.ItemDataRole.UserRole, model.name)
            if model.client in icons:
                model_item.setIcon(icons[model.client])
            self.list_widget.addItem(model_item)

    def get_selected_models(self):
        """
        Get the selected models.
        """
        if not self.list_widget.selectedItems():
            return ""
        return ",".join([item.data(Qt.ItemDataRole.UserRole) for item in self.list_widget.selectedItems()])

    def set_selected_models(self, models: str):
        """
        Set the selected models.
        :param models: The models to select.
        """
        models_list = models.split(",") if models else []
        for item in self.list_widget.findItems("", Qt.MatchFlag.MatchContains):
            if item.data(Qt.ItemDataRole.UserRole) in models_list:
                item.setSelected(True)

    models = Property(str, get_selected_models, set_selected_models)

    def open_context_menu(self, position):
        """
        Open the context menu at the specified position.
        :param position: The position to open the context menu.
        """
        context_menu = QMenu()
        action = context_menu.addAction("Clear selection")
        action.triggered.connect(lambda: self.list_widget.clearSelection())
        context_menu.exec(self.mapToGlobal(position))

    def on_selection_changed(self):
        """
        Handler for when the selection changes.
        """
        selected_models = [
            item.data(Qt.ItemDataRole.UserRole) for item in self.list_widget.selectedItems()
        ]
        self.selected_models_label.setText(', '.join(selected_models))
        self.selected_models_label.adjustSize()
        self.layout.addWidget(self.selected_models_label)



from quack2tex.pyqt import (
    Qt, QPoint, QRect, QSize,QDialog, QRubberBand
)

class ScreenCaptureWindow(QDialog):
    """
    A semi-transparent window that allows the user to select a region to capture
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Select Region to Capture")
        self.setWindowOpacity(0.3)
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.start_point = QPoint()
        self.end_point = QPoint()
        self.rubber_band = QRubberBand(QRubberBand.Shape.Rectangle, self)
        self.selected_region = None

    def mousePressEvent(self, event):
        """
        Start the rubber band selection on mouse press
        :param event:
        :return:
        """
        self.start_point = event.pos()
        self.rubber_band.setGeometry(QRect(self.start_point, QSize()))
        self.rubber_band.show()

    def mouseMoveEvent(self, event):
        """
        Update the rubber band selection on mouse move
        :param event:
        :return:
        """
        self.rubber_band.setGeometry(QRect(self.start_point, event.pos()).normalized())

    def mouseReleaseEvent(self, event):
        """
        Capture the selected region on mouse release
        :param event:
        :return:
        """
        self.end_point = event.pos()
        self.selected_region = self.capture_region()
        self.close()

    def keyPressEvent(self, event):
        """
        Close the window on escape key press
        :param event:
        :return:
        """
        if event.key() == Qt.Key.Key_Escape:
            self.selected_region = None
            self.close()

    def capture_region(self):
        """
        Capture the selected region
        :return:
        """
        x1 = min(self.start_point.x(), self.end_point.x())
        y1 = min(self.start_point.y(), self.end_point.y())
        x2 = max(self.start_point.x(), self.end_point.x())
        y2 = max(self.start_point.y(), self.end_point.y())

        # # relative to the screen
        # left_percent = x1 / self.screen().geometry().width()
        # top_percent = y1 / self.screen().geometry().height()
        # width_percent = (x2 - x1) / self.screen().geometry().width()
        # height_percent = (y2 - y1) / self.screen().geometry().height()

        return (x1, y1, x2 - x1, y2 - y1)



from .session_manager import SessionManager
from quack2tex.utils import LibUtils
from sqlalchemy.pool import NullPool
from sqlalchemy.orm import Session


db_sync_connection_string, _ = LibUtils.get_db_connection_string()
sessionmanager = SessionManager(
    url=db_sync_connection_string,
    async_mode=False,
    engine_kwargs={"echo": False, "poolclass": NullPool},
)

def get_db_session(*args, **kwargs) -> Session:
    """
    Get database session
    """
    drop_all = kwargs.pop("drop_all", False)
    sessionmanager.init(drop_all=drop_all)
    return sessionmanager.session(*args, **kwargs)


from quack2tex.pyqt import (
    Signal, QSize, QIcon, QPushButton
)


class ImageButton(QPushButton):
    """
    A custom QPushButton that emits a doubleClicked signal when double clicked
    """

    doubleClicked = Signal()

    def __init__(self, image_path: str, icon_size: QSize, parent=None):
        super().__init__(parent)
        self.setAutoFillBackground(True)
        self.setStyleSheet("""
            QPushButton {
                background-color: transparent;  /* Transparent background */
                border: none;                    /* Remove border */
                padding: 10px;
                color: black;                    /* Set text color */
            }
            
            QPushButton:pressed {
                background-color: transparent;  /* Transparent when pressed */
            }
        
            QPushButton:hover {
                background-color: transparent;  /* Transparent on hover */
            }
        """)
        self.setContentsMargins(0, 0, 0, 0)
        self.image_path = image_path
        self.setIconSize(icon_size)
        self.setIcon(QIcon(image_path))
        #self.setFlat(True)
        self.setFixedSize(icon_size)

    def sizeHint(self):
        """
        Return the size hint
        :return: QSize object with the size hint
        """
        return self.iconSize()


    def mouseDoubleClickEvent(self, event):
        """
        Emit the doubleClicked signal on double click
        :param event:
        :return:
        """
        self.doubleClicked.emit()



from quack2tex.pyqt import Qt,QStandardItemModel, QStandardItem


class TreeViewStandardItemModel(QStandardItemModel):
    """
    Custom model class based on QStandardItemModel.
    """

    def __init__(self, columns):
        """
        Initialize the custom model with the given columns.
        """
        super().__init__(0, len(columns))  # Initialize model with given number of columns
        self._columns = columns
        self.setHorizontalHeaderLabels(columns)

    def reset_model(self):
        """
        Reset the model.
        """
        self.clear()
        self.setHorizontalHeaderLabels(self._columns)

    def flags(self, index):
        """
        Return the flags for the given index.
        """
        flags = super().flags(index)
        flags &= ~Qt.ItemFlag.ItemIsEditable  # Corrected flag usage
        return flags

    def remove_checked_items(self):
        """
        Remove all items that are checked.
        """
        def recursive_remove_items(parent: QStandardItem):
            """
            Recursively check all children and remove checked items.
            """
            rows_to_remove = []
            for row in range(parent.rowCount()):
                child_item = parent.child(row)
                # Check if the item is checked
                if child_item.checkState() == Qt.CheckState.Checked:  # Corrected check state usage
                    rows_to_remove.append(row)
                else:
                    # Recursively check the item's children
                    recursive_remove_items(child_item)
            # Remove rows from the bottom to the top to prevent index issues
            for row in reversed(rows_to_remove):
                parent.removeRow(row)

        # Start the process from the root items in the model
        root = self.invisibleRootItem()
        recursive_remove_items(root)

    def check_all_items(self):
        """
        Check all items in the tree.
        """
        def recursive_check_items(parent: QStandardItem):
            for row in range(parent.rowCount()):
                child_item = parent.child(row)
                child_item.setCheckState(Qt.CheckState.Checked)  # Corrected check state usage
                recursive_check_items(child_item)

        root = self.invisibleRootItem()
        recursive_check_items(root)

    def uncheck_all_items(self):
        """
        Uncheck all items in the tree.
        """
        def recursive_uncheck_items(parent: QStandardItem):
            for row in range(parent.rowCount()):
                child_item = parent.child(row)
                child_item.setCheckState(Qt.CheckState.Unchecked)  # Corrected check state usage
                recursive_uncheck_items(child_item)

        root = self.invisibleRootItem()
        recursive_uncheck_items(root)

    def add_child(self, data, parent_item=None):
        """
        Add a child item with the given data.
        """
        new_item = QStandardItem(data)
        new_item.setCheckable(True)

        if parent_item is None:
            # Add as a top-level item
            self.appendRow(new_item)
        else:
            # Add as a child to the given parent item
            parent_item.appendRow(new_item)

    def find_checked_nodes_recursive(self, parent_item=None):
        """
        Find all checked items in the model.
        """
        checked_items = []
        if parent_item is None:
            parent_item = self.invisibleRootItem()

        for row in range(parent_item.rowCount()):
            child_item = parent_item.child(row)
            if child_item.checkState() == Qt.CheckState.Checked:  # Corrected check state usage
                checked_items.append(child_item)
            checked_items += self.find_checked_nodes_recursive(child_item)

        return checked_items

    def set_data(self, index, value, role):
        """
        Set the data for the given index and role.
        """
        item = self.itemFromIndex(index)
        if item:
            if role == Qt.ItemDataRole.CheckStateRole:  # Corrected role usage
                item.setCheckState(Qt.CheckState.Checked if value else Qt.CheckState.Unchecked)  # Corrected check state usage
                return True
            elif role == Qt.ItemDataRole.EditRole:  # Corrected role usage
                item.setText(value)
                return True
        return False

    def refresh_model(self):
        """
        Refresh the model layout.
        """
        self.layoutChanged.emit()


import os
from pathlib import Path

class LibUtils:
    """
    Library utilities
    """

    @classmethod
    def get_lib_home(cls):
        """
        it returns the path to the library home
        :return:
        """

        default_folder = Path.home() / ".quack2tex"
        lib_home = Path(os.getenv("QUACK_HOME", default_folder))
        lib_home.mkdir(parents=True, exist_ok=True)
        return lib_home

    @classmethod
    def get_db_connection_string(cls):
        """
        Get the database and artifacts location
        :return:
        """
        lib_home = cls.get_lib_home()
        db_file = lib_home / "quack2tex.db"
        default_db_sync_connection_string = f"sqlite:///{db_file}"
        default_db_async_connection_string = f"sqlite+aiosqlite:///{db_file}"
        db_async_connections_string = os.getenv(
            "QUACK2TEX_ASYNC_DB_CONNECTION_STRING", default_db_async_connection_string
        )
        db_sync_connections_string = os.getenv(
            "QUACK2TEX_SYNC_DB_CONNECTION_STRING", default_db_sync_connection_string
        )
        return db_sync_connections_string, db_async_connections_string



from quack2tex.pyqt import QToolBox, QDialog, QVBoxLayout

from quack2tex.widgets import MarkdownViewer


class OutputDialog(QDialog):
    """
    A window to display the predictions output
    """
    def __init__(self, predictions: dict, parent=None):
        super().__init__(parent)

        # Set window properties
        self.setWindowTitle("Matrix Theme Window")
        self.setGeometry(100, 100, 800, 600)
        toolbox = QToolBox()
        for model, prediction in predictions.items():
            viewer = MarkdownViewer()
            viewer.set_content(prediction)
            toolbox.addItem(viewer, model)

        self.layout = QVBoxLayout()
        self.layout.addWidget(toolbox)
        self.setLayout(self.layout)




import math

from quack2tex.pyqt import (
    QGraphicsDropShadowEffect,
    QColor,
    QTimer,
    QSize,
    QParallelAnimationGroup,
    QGraphicsOpacityEffect,
    QPropertyAnimation,
    QPoint,
    QEasingCurve,
)
from quack2tex.widgets import ImageButton


class FloatingMenuItem(ImageButton):
    """
    A floating menu item that displays an icon and can have children items.
    """

    def __init__(self, icon_path: str, icon_size: QSize, distance_to_center=100, start_angle=0, end_angle=360, root=None, parent=None, data=None):
        super().__init__(icon_path, icon_size, parent)
        self.distance_to_center = distance_to_center
        self.start_angle = start_angle
        self.end_angle = end_angle
        self.children = []
        self.root = root
        self.clicked.connect(self.addGlowEffect)
        self.data = data
        self._expanded = False
        self._collapsed = True

    @property
    def data(self):
        return self.property("data")

    @data.setter
    def data(self, data):
        self.setProperty("data", data)


    def addGlowEffect(self):
        """
        Add a glow effect to the button
        :return:
        """
        # Create a shadow effect to simulate the glow
        glow = QGraphicsDropShadowEffect(self)
        glow.setBlurRadius(40)  # Increase blur radius for a strong glow effect
        glow.setColor(QColor(255, 255, 224))  # Neon cyan color (RGB: 0, 255, 255)
        glow.setOffset(0, 0)  # No offset, straight glow around the button
        # Apply the effect to the button
        self.setGraphicsEffect(glow)
        # Remove the glow after 1 second
        QTimer.singleShot(1000, self.removeGlowEffect)

    def removeGlowEffect(self):
        """
        Remove the glow effect from the button
        :return:
        """
        # Remove the shadow effect, returning the button to its normal state
        self.setGraphicsEffect(None)

    def fade(self):
        """
        Fade the button
        :return:
        """
        self.effect = QGraphicsOpacityEffect()
        self.setGraphicsEffect(self.effect)
        self.animation = QPropertyAnimation(self.effect, b"opacity")
        self.animation.setDuration(300)
        self.animation.setStartValue(1)
        self.animation.setEndValue(0)
        self.animation.start()

    def unfade(self):
        """
        Unfade the button
        :return:
        """

        self.effect = QGraphicsOpacityEffect()
        self.setGraphicsEffect(self.effect)
        self.animation = QPropertyAnimation(self.effect, b"opacity")
        self.animation.setDuration(300)
        self.animation.setStartValue(0)
        self.animation.setEndValue(1)
        self.animation.start()

    def add_child(self, child: "FloatingMenuItem"):
        """
        Add a child item to the menu item.
        :param child:  the child item to add
        :return:
        """
        self.children.append(child)

    def is_collapsed(self) -> bool:
        """
        Returns True if all children are hidden (collapsed).
        :return:
        """
        #return any(not child.isVisible() for child in self.children)
        return self._collapsed

    def is_expanded(self) -> bool:
        """
        Returns True if at least one child is visible (expanded).
        :return:
        """
        #return any(child.isVisible() for child in self.children)
        return self._expanded

    def get_items_angles(self, num_items):
        """
        Get the angles for the items
        :param num_items: number of items
        :return: list of angles
        """
        angle_range = self.end_angle - self.start_angle
        full_menu = self.start_angle == 0 and self.end_angle == 360
        angle_increment = angle_range / (
            num_items if full_menu else max(1, num_items - 1)
        )
        angles_rad = [math.radians(self.start_angle + i * angle_increment) for i in range(num_items)]
        return angles_rad

    def expand(self):
        """
        Expand the item and show its children with animation.
        :return:
        """
        if self.is_expanded() or not self.children:
            return

        self.raise_()
        center = self.geometry().center()
        num_children = len(self.children)
        angles_rad = self.get_items_angles(num_children)
        animation_group = QParallelAnimationGroup(self)
        for i, child in enumerate(self.children):
            child.move(QPoint(center.x() - child.width() // 2, center.y() - child.height() // 2))
            x = math.ceil(center.x() + self.distance_to_center * math.cos(angles_rad[i]) - child.width() // 2)
            y = math.ceil(center.y() + self.distance_to_center * math.sin(angles_rad[i]) - child.height() // 2)
            child.show()

            anim = QPropertyAnimation(child, b"pos")
            anim.setEndValue(QPoint(x, y))
            anim.setDuration(500)
            anim.setEasingCurve(QEasingCurve.Type.OutBounce)
            animation_group.addAnimation(anim)
            animation_group.finished.connect(lambda: self.expand_animation_group_finished(child))

        animation_group.start()

    def expand_animation_group_finished(self, child):
        """
        Handle the animation group finished event
        :return:
        """
        self._expanded = True
        self._collapsed = False
        self.raise_()



    def collapse(self):
        """
        Collapse the item and hide its children with animation.
        :return:
        """

        if self.is_collapsed() or not self.children:
            return

        center = self.geometry().center()
        animation_group = QParallelAnimationGroup(self)

        for child in self.children:
            anim = QPropertyAnimation(child, b"pos")
            anim.setEndValue(QPoint(center.x() - child.width() // 2, center.y() - child.height() // 2))
            anim.setDuration(500)
            anim.setEasingCurve(QEasingCurve.Type.OutBounce)
            anim.finished.connect(lambda: self.collapse_animation_group_finished(child))
            animation_group.addAnimation(anim)

        animation_group.start()

    def collapse_animation_group_finished(self, child):
        """
        Handle the animation group finished event
        :return:
        """
        child.hide()
        self._expanded = False
        self._collapsed = True


    def hide_children(self, parent_item: "FloatingMenuItem"):
        """
        Hide the children of a given item.
        :param parent_item:
        :return:
        """
        try:
            for child in parent_item.children:
                child.hide()
                if child.children:
                    self.hide_children(child)
        except Exception as e:
            print(e)

    def collapse_siblings(self):
        """
        Collapse the siblings of this item.
        :return:
        """
        center = self.root.geometry().center()
        group = QParallelAnimationGroup(self)
        siblings = [child for child in self.root.children if child != self]
        for sibling in siblings:
            anim = QPropertyAnimation(sibling, b"pos")
            anim.setEndValue(QPoint(center.x() - sibling.width() // 2, center.y() - sibling.height() // 2))
            anim.setDuration(500)
            anim.setEasingCurve(QEasingCurve.Type.OutBounce)
            anim.finished.connect(sibling.hide)
            group.addAnimation(anim)
        group.start()

    def expand_siblings(self):
        """
        Expand the siblings of this item.
        :return:
        """
        root_center = self.root.geometry().center()
        num_siblings = len(self.root.children)
        angles_rad = self.get_items_angles(num_siblings)

        for i, sibling in enumerate(self.root.children):
            if sibling == self:
                continue
            x = math.ceil(root_center.x() + self.root.distance_to_center * math.cos(angles_rad[i]) - sibling.width() // 2)
            y = math.ceil(root_center.y() + self.root.distance_to_center * math.sin(angles_rad[i]) - sibling.height() // 2)
            sibling.move(x, y)
            sibling.setVisible(True)
            sibling.unfade()

    def mousePressEvent(self, e):
        """
        Triggered when the user presses the mouse button
        :param e:
        :return:
        """
        # Propagate the event to the parent window
        if self.window() is not None:
            self.window().mousePressEvent(e)
        super().mousePressEvent(e)

    def toggle(self):
        """
        Toggle the item between expanded and collapsed states.
        :return:
        """
        if self.is_collapsed():
            self.handle_expand()
        else:
            self.handle_collapse()

    def handle_expand(self):
        """Handle the logic for expanding the item."""
        if self.children:
            if self.root:
                self.collapse_siblings()
            self.expand()
            self.raise_()

    def handle_collapse(self):
        """Handle the logic for collapsing the item."""
        if self.children:
            for child in self.children:
                self.hide_children(child)
            self.collapse()
            self.raise_()

        if self.root:
            if self.is_expanded():
                self.expand_siblings()



